"""Housie Ticket generation logic"""

__author__ = 'Aaron Alphonso'
__email__ = 'alphonsoaaron1993@gmail.com'

from collections import defaultdict
from random import choice, randint
from typing import List, Set, Dict, DefaultDict

from housie.constants import Number, Row, ColumnRange, COLUMN_RANGES
from housie.models import Ticket


def generate_ticket() -> Ticket:
    """Generates a Housie Ticket containing 15 randomly selected numbers based on the following rules

    * A Housie ticket has 15 numbers.
    * These 15 numbers are placed in a 3 rows x 9 cols grid (27 possible spaces).
    * There has to be exactly 5 numbers in each row.
    * There has to be at least one number in each column.
    * The numbers in each column must belong to that column's allowed range.
        Column 1: 1-9
        Column 2: 10-19
        Column 3: 20-29
        .      .      .
        .      .      .
        .      .      .
        Column 8: 70-79
        Column 9: 80-90

    Based on the rules above we have to select the numbers.
    """
    # The list of ranges for each column. Make a copy as we mutate it during our selection process
    column_ranges: List[ColumnRange] = COLUMN_RANGES.copy()

    # The numbers selected for the ticket
    ticket_numbers: Set[Number] = set()

    # Saving the numbers column-wise to ensure we select at max 3 numbers for each column and also
    # to help in distributing them into the 3 rows later
    column_wise_numbers: DefaultDict[ColumnRange, List[Number]] = defaultdict(list)

    # Select one number from each column range. This will give us 9 numbers
    for column_range in column_ranges:
        selected_number = select_unique_number_from_range(
            column_range=column_range, already_selected=column_wise_numbers[column_range])
        ticket_numbers.add(selected_number)
        column_wise_numbers[column_range].append(selected_number)

    # Select the remaining 6 numbers at random from the columns.
    while len(ticket_numbers) < 15:
        selected_range = choice(column_ranges)
        selected_number = select_unique_number_from_range(
            column_range=selected_range, already_selected=column_wise_numbers[selected_range])

        ticket_numbers.add(selected_number)
        column_wise_numbers[selected_range].append(selected_number)

        # If we have selected 3 numbers for a specific column, remove that column from the further selection process
        if len(column_wise_numbers[selected_range]) == 3:
            column_ranges.remove(selected_range)

    # Convert to a regular dict now as we no longer need the defaultdict properties
    # Now we assign the 15 selected numbers into the 3 rows.
    rows = assign_to_rows(dict(column_wise_numbers))
    return Ticket(rows)


def select_unique_number_from_range(column_range: ColumnRange, already_selected: List[Number]) -> Number:
    """Selects and returns a number randomly from a ColumnRange. If the number has already been selected before, then
    re-select a new number"""
    selected_number: Number = randint(column_range.start, column_range.end)
    # If the selected_number is already_selected, select a new number until we get a unique number
    while selected_number in already_selected:
        selected_number = randint(column_range.start, column_range.end)
    return selected_number


def assign_to_rows(column_wise_numbers: Dict[ColumnRange, List[Number]]) -> List[Row]:
    """Distribute the ticket numbers into the 3 rows such that each row has exactly 5 numbers

    Start with the columns that have 3 numbers, followed by the columns that have two numbers and
    then columns that have 1 number.
    Assign the numbers in a round-robin fashion.
    """
    ticket: List[Row] = [[], [], []]
    row_num = 0
    for col_len in [3, 2, 1]:
        columns = list(filter(lambda x: len(x) == col_len, column_wise_numbers.values()))
        for column in columns:
            row_num = insert_from_columns_to_rows(column, ticket, row_num)
    for row in ticket:
        row.sort()
    return ticket


def insert_from_columns_to_rows(column: List[Number], ticket: List[Row], row_num: int) -> int:
    """Insert numbers from the columns into the rows in a round-robin fashion.
    Returns the row_num, so it can be persisted between calls"""
    numbers = sorted(column)

    # If we have to assign numbers to row 3 and row 1 of the same column, we should always assign the smaller
    # number to the higher row, i.e. row 1. To do this, we have the following logic:
    # Figure out which rows we will be assigning the column numbers to. Assign in round robin logic starting from the
    # smallest row num possible.
    row_nums_to_assign_to = []
    for _ in numbers:
        row_nums_to_assign_to.append(row_num)
        row_num = (row_num + 1) % 3

    # This is the last row position that we are assigning to in the round-robin assignment. Return this for the
    # assignment of the numbers from the next column
    last_row_num = row_num
    row_nums_to_assign_to.sort()
    for number, row_num in zip(numbers, row_nums_to_assign_to):
        ticket[row_num].append(number)
    return last_row_num


def demo_ticket_generation() -> None:
    """Demonstrates ticket generation by generating a ticket and printing it visually"""
    sample_ticket = generate_ticket()
    print("Concise representation")
    print(sample_ticket.minimalistic_display())

    print("Structural representation")
    print(sample_ticket.structural_display())



"""Simple Housie Ticket class to store and display a ticket"""

__author__ = 'Aaron Alphonso'
__email__ = 'alphonsoaaron1993@gmail.com'

from typing import List, Dict, Optional, Set

from housie.constants import Row, COLUMN_RANGES, Number, NUMBER_POOL
from housie.utils import load_json, strike_through, clear_screen


class Ticket:
    """Simple structure to store and display the ticket"""

    def __init__(self, rows: List[Row]):
        # self.rows will hold only the non-None elements
        self.rows: List[Row] = []
        # Storing all the numbers in a flat list
        self.numbers: List[int] = []
        for row in rows:
            new_row = []
            for number in row:
                if number:
                    self.numbers.append(number)
                    new_row.append(number)
            self.rows.append(new_row)
        # All the numbers that are selected
        self.selected: Set[Number] = set()

    def mark_number(self, number: int) -> None:
        """Updates the ticket marking the matching number as selected"""
        if number in self.numbers:
            self.selected.add(number)

    def mark_numbers(self, numbers: List[int]) -> None:
        """Updates the ticket marking the matching numbers as selected"""
        for number in numbers:
            self.mark_number(number)

    def __repr__(self) -> str:
        return "Ticket(numbers={})".format(self.rows)

    def display_ticket(self) -> str:
        """Display the ticket"""
        return self.structural_display()

    def minimalistic_display(self) -> str:
        """Displays the ticket in a minimalistic format"""
        ticket_representation = ''
        for row in self.rows:
            row_elems = []
            for number in row:
                if number:
                    if number in self.selected:
                        row_elems.append(strike_through('{: 3} '.format(number)))
                    else:
                        row_elems.append('{: 3} '.format(number))
            ticket_representation += ','.join(row_elems) + '\n'
        return ticket_representation

    def structural_display(self) -> str:
        """Displays the ticket in a 3 x 9 grid as is seen on traditional tickets"""
        ticket_representation = ''
        for row in self.get_structural_representation():
            for num in row:
                if num:
                    if num in self.selected:
                        ticket_representation += strike_through('{: 3} '.format(num))
                    else:
                        ticket_representation += '{: 3} '.format(num)
                else:
                    ticket_representation += '  - '
            ticket_representation += '\n'
        return ticket_representation

    def get_structural_representation(self) -> List[List[Optional[Number]]]:
        """Returns a representation of the ticket in a 3 x 9 grid with blank squares represented by None

        This representation can be used for a graphical display.
        """
        representation = []
        for row in self.rows:
            new_row: List[Optional[Number]] = []
            row_iter = iter(row)
            num = next(row_iter)
            if num:
                for start, end in COLUMN_RANGES:
                    if num and start <= num <= end:
                        new_row.append(num)
                        try:
                            num = next(row_iter)
                        except StopIteration:
                            # Add 'None' to the remaining elements of the row
                            new_row.extend([None] * (9 - len(new_row)))
                            break
                    else:
                        new_row.append(None)
                representation.append(new_row)
        return representation


def load_tickets(file_name: str) -> Optional[Dict[str, List[Ticket]]]:
    """Reads the input file and tries to parse the data into a dict of name : tickets"""
    ticket_data = load_json(file_name)
    if not ticket_data:
        return None
    for name, tickets in ticket_data.items():
        tickets = list(map(Ticket, tickets))
        ticket_data[name] = tickets
    return ticket_data


def demo_ticket() -> None:
    """Method to demonstrate how tickets would look like with some numbers marked"""
    from housie.generate_ticket import generate_ticket

    # Generate some sample tickets
    names = ['Thor', 'Ironman', 'Spiderman']
    ticket_data = {}
    for name in names:
        ticket = generate_ticket()
        ticket_data[name] = [ticket]

    # Mark some random numbers so we can see the output
    from random import sample
    random_numbers = sample(NUMBER_POOL, 40)
    for name, tickets in ticket_data.items():
        for ticket in tickets:
            for number in random_numbers:
                ticket.mark_number(number)

    # Display the tickets
    clear_screen()
    for name, tickets in ticket_data.items():
        print(name)
        for ticket in tickets:
            print(ticket.display_ticket())



"""Model representing the Housie board"""

__author__ = 'Aaron Alphonso'
__email__ = 'alphonsoaaron1993@gmail.com'

from typing import List, Optional
from random import choice

from housie.constants import SelectedPool, RemainingPool, Number, NUMBER_POOL
from housie.utils import clear_screen


class Board:
    """Represents the Housie board state and logic to run the game and display the state"""

    def __init__(self, already_selected: Optional[SelectedPool] = None) -> None:
        """Initialize the state of the board.
        Either a blank new board, or some pre-selected numbers to represent an already mid-way game"""
        # Variables for holding game state
        self.remaining: RemainingPool = []
        self.selected: SelectedPool = []
        if already_selected:
            self.init_custom_game(already_selected)
        else:
            self.init_new_game()

    def pick_next(self) -> Optional[Number]:
        """Picks a number at random from the remaining pool of numbers.
        The chosen number is removed from the remaining_pool and added to the selected pool"""
        if not self.remaining:
            return None
        number = choice(self.remaining)
        return self.pick_manual(number)

    def pick_many(self, count: int) -> List[Number]:
        """Pick multiple numbers from the remaining pool of numbers
        The chosen numbers are removed from the remaining pool and added to the selected pool"""
        return list(filter(None, [self.pick_next() for i in range(count)]))

    def pick_manual(self, number: Number) -> Number:
        """Manually pick a number from the number pool.
        The chosen number is removed from the remaining pool and added to the selected pool"""
        if number in self.remaining:
            self.remaining.remove(number)
            self.selected.append(number)
        return number

    def init_new_game(self) -> None:
        """Initialize a new game"""
        self.selected = []
        self.remaining = NUMBER_POOL.copy()

    def init_custom_game(self, already_selected: List[Number]) -> None:
        """Initialize a new game with a set of numbers already called out"""
        self.selected = already_selected
        self.remaining = list(set(NUMBER_POOL) - set(already_selected))

    def last_5_selected(self) -> List[Number]:
        """Returns a list of the last 5 numbers called out"""
        return self.selected[-5:]

    def display_board(self) -> str:
        """Display the Housie Board of selected numbers visually in the console"""
        board = ''
        board += '------------------ Housie Board -------------------\n'
        row_size = 10
        num_rows = len(NUMBER_POOL) // row_size
        for row_num in range(num_rows):
            start = row_num * row_size
            end = (row_num * row_size) + row_size
            for num in NUMBER_POOL[start: end]:
                if num in self.selected:
                    board += '|{: 3} '.format(num)
                else:
                    board += '|' + ' ' * 4
            board += '|'
            board += '\n---------------------------------------------------\n'

        last_5_numbers_str = ", ".join(map(str, reversed(self.last_5_selected())))
        board += 'Last 5 Numbers called: {}\n'.format(last_5_numbers_str)
        board += 'Left in the bag: {}\n'.format(len(self.remaining))
        return board


def demo_board() -> None:
    """Test function to print out how the board would look with a few numbers filled"""
    housie_board = Board()
    housie_board.pick_many(40)
    clear_screen()
    print(housie_board.display_board())



"""Complex logic for displaying a followed game conveniently in terminal"""

__author__ = 'Aaron Alphonso'
__email__ = 'alphonsoaaron1993@gmail.com'

from itertools import zip_longest
from typing import Dict, List

from housie.models import Board, Ticket


def display_followed_game(board: Board, ticket_data: Dict[str, List[Ticket]]) -> None:
    """Use this to select how you want the game displayed. The complex display looks better, but may not work in all
    sizes of terminals. Use the simple display as a fall back"""
    # Side-by-Side display with minimalistic ticket design
    # _complex_display_followed_game(housie, ticket_data, 'minimalistic_display', 24)

    # Side-by-Side display with traditional/structural ticket design
    _complex_display_followed_game(board, ticket_data, 'structural_display', 36)


# Simplistic vertical scrolling display
# _simple_display_followed_game(housie, ticket_data)


def _simple_display_followed_game(board: Board, ticket_data: Dict[str, List[Ticket]]) -> None:
    """Simple Display method for use in following a game

    Displays the board followed by the tickets
    """
    print(board.display_board())
    for name, tickets in ticket_data.items():
        print(name)
        for ticket in tickets:
            print(ticket.display_ticket())


class DisplayElement:
    """A wrapper object to group together multiple lines of text as a single element"""

    def __init__(self, content: str) -> None:
        self.row_wise_data = content.split('\n')
        self.content = content


def _complex_display_followed_game(
        board: Board, ticket_data: Dict[str, List[Ticket]], display_method: str, line_len: int) -> None:
    """Side-by-Side Display method for use in following a game.
    Tries to fit the tickets besides the board instead of below.

    Not the best implementation though, as it involves a lot of string manipulations.
    Also doesn't consider size of the terminal for now. But it works!

    Allows to display up to 8 tickets in a nice clean manner. After that, the view could get a little messy
    """
    board_element = DisplayElement(board.display_board())
    ticket_elements = []
    for name, tickets in ticket_data.items():
        ticket_str = ''
        for ticket in tickets:
            ticket_str += getattr(ticket, display_method)() + '\n'
        ticket_str = name + (line_len - len(name)) * '-' + '\n' + ticket_str
        ticket_elements.append(DisplayElement(ticket_str))

    if len(ticket_elements) > 1:
        combined_element = fold(ticket_elements)
        further_combined = fold([board_element, combined_element])
        print(further_combined.content)
    else:
        combined_element = fold([board_element, ticket_elements[0]])
        print(combined_element.content)


def fold(display_elements: List[DisplayElement]) -> DisplayElement:
    """Split the incoming list of elements into two side-by-side columns"""
    middle = len(display_elements) // 2
    column1 = display_elements[:middle]
    column2 = display_elements[middle:]

    # Flatten the List of display elements into a list of rows. (each display element contains rows)
    column1_rows = []
    for element in column1:
        for row in element.row_wise_data:
            column1_rows.append(row)

    column2_rows = []
    for element in column2:
        for row in element.row_wise_data:
            column2_rows.append(row)

    combined_str = ''
    space = ' '
    fill_value = len(column1_rows[0]) * ' ' if len(column2_rows) > len(column1_rows) else len(column2_rows[0]) * ' '
    for col1, col2 in zip_longest(column1_rows, column2_rows, fillvalue=fill_value):
        combined_str += col1 + 5 * space + col2 + '\n'
    return DisplayElement(combined_str)
